<?php defined('SYSPATH') or die('No direct script access.');

/**
 * @property-read $ID int
 */
abstract class Model_Container extends Model implements JsonSerializable, ArrayAccess
{
    protected static $id_name = 'ID';

    protected $data;

    /**
     * @var bool
     * Indicates whether data was loaded from db.
     */
    protected $loaded = false;

    /**
     * @var bool
     * Allow to load data from database, if current
     * object instance does not contains some fields.
     */
    protected $lazy = true;

    /**
     * @param $data array
     * Source data for building the object.
     * @param $associated bool
     * Indicates whether object exists in db.
     * @param $lazyLoading bool
     * Indicates whether lazy loading can be used.
     */
    public function __construct(array & $data, $associated = true, $lazyLoading = true) {
        $this->data = $data;
        $this->lazy = $lazyLoading;
        if (!$associated || !isset($data[self::$id_name]))
            $this->create();
    }

    /**
     * Creation of new object from raw data.
     */
    static public function make() { }

    /**
     * Load the object from database.
     * @param $id int  object id
     * @return $this  new instance
     */
    static public function load($id) {
        if (!is_numeric($id) || $id <= 0)
            throw new InvalidArgumentException(Error::ID_IS_INCORRECT);

        // lazy loading
        $data[self::$id_name] = (int) $id;
        return new static($data, true);
    }

    /**
     * Create new object in db, based on $data.
     */
    protected abstract function create();

    public function __get($name) {
        $name = ucfirst($name);

        // if value has already been loaded
        if (array_key_exists($name, $this->data))
            return $this->data[$name];

        // if no, then make query to db
        $this->loadData();
        if (array_key_exists($name, $this->data))
            return $this->data[$name];

        throw new ErrorException('No such field or property');
    }

    public function __set($name, $value) {
        if ($name == self::$id_name)
            throw new BadMethodCallException();
        $this->data[$name] = $value;
    }

    public function offsetExists($name) {
        $name = ucfirst($name);

        if (array_key_exists($name, $this->data))
            return true;

        $this->loadData();
        return array_key_exists($name, $this->data);
    }

    public function offsetGet($offset) {
        return $this->__get($offset);
    }

    public function offsetSet($offset, $value) {
        $this->__set($offset, $value);
    }

    public function offsetUnset($offset) {
        throw new BadMethodCallException();
        // this method may be confused by lazy loading
    }

    /** @deprecated */
    public function toArray() {
        $this->loadData();
        return $this->data;
    }

    public function jsonSerialize($forceLoad = true) {
        if ($forceLoad)
            $this->loadData();
        return $this->data;
    }

    protected function loadData() {
        if (!$this->loaded && $this->lazy) {
            $this->data = array_merge($this->data, static::getInfo($this->ID));
            $this->loaded = true;
        }
    }

    /**
     * @param $id int object id
     * @return array data from database
     */
    protected abstract function getInfo($id);
}